import numpy as np

from itertools import chain, combinations

import json
import re

import nshap


def allclose(dict1, dict2, rtol=1e-05, atol=1e-08):
    """Compare if two dicts of n-Shapley Values are close according to numpy.allclose.

    Useful for testing purposes.

    Args:
        dict1 (dict): The first dict.
        dict2 (dict): The second dict.
        rtol (float, optional): passed to numpy.allclose. Defaults to 1e-05.
        atol (float, optional): passed to numpy.allclose. Defaults to 1e-08.

    Returns:
        bool: Result of numpy.allclose.
    """
    if dict1.keys() != dict2.keys():  # both dictionaries need to have the same keys
        return False
    for key in dict1.keys():
        if not np.allclose(dict1[key], dict2[key], rtol=rtol, atol=atol):
            return False
    return True


def save(values, fname):
    """Save an interaction index to a JSON file.

    Args:
        values (nshap.InteractionIndex): The interaction index.
        fname (str): Filename.
    """
    json_dict = {
        str(k): v for k, v in values.data.items()
    }  # convert the integer tuples to strings
    json_dict["index_type"] = values.index_type
    json_dict["n"] = values.n
    json_dict["d"] = values.d

    with open(fname, "w+") as fp:
        fp.write(json.dumps(json_dict, indent=2))


def to_int_tuple(str_tuple):
    """Convert string representations of integer tuples back to python integer tuples.

    This utility function is used to load n-Shapley Values from JSON.

    Args:
        str_tuple (str): String representaiton of an integer tuple, for example "(1,2,3)"

    Returns:
        tuple: Tuple of integers, for example (1,2,3)
    """
    start = str_tuple.find("(") + 1
    first_comma = str_tuple.find(",")
    end = str_tuple.rfind(")")
    # is the string of the form "(1,)", i.e. with a trailing comma?
    if len(re.findall("[0-9]", str_tuple[first_comma:])) == 0:
        end = first_comma
    str_tuple = str_tuple[start:end]
    return tuple(map(int, str_tuple.split(",")))


def read_remove(d, key):
    """Read and remove a key from a dict d.
    """
    r = d[key]
    del d[key]
    return r


def load(fname):
    """Load an interaction index from a JSON file.

    Args:
        fname (str): Filename.

    Returns:
        nshap.InteractionIndex: The loaded interaction index.
    """
    with open(fname, "r") as fp:
        str_dump = json.load(fp)

    index_type = read_remove(str_dump, "index_type")
    n = read_remove(str_dump, "n")
    d = read_remove(str_dump, "d")

    # convert the remaining string tuples to integer tuples
    python_dict = {to_int_tuple(k): v for k, v in str_dump.items()}
    return nshap.InteractionIndex(index_type, python_dict, n, d)


def powerset(iterable):
    "The powerset function from https://docs.python.org/3/library/itertools.html#itertools-recipes."
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))



import numpy as np

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

import seaborn as sns

import nshap

# avoid type-3 fonts
matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42

#############################################################################################################################
#                                                       Plots
#############################################################################################################################

plot_colors = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
    "black",
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]


def plot_interaction_index(
    I: nshap.InteractionIndex,
    max_degree=None,
    axis=None,
    feature_names=None,
    rotation=70,
    legend=True,
    fig_kwargs={"figsize": (6, 6)},
    barwidth=0.5,
):
    """Generate the plots in the paper.

    Args:
        n_shapley_values (nshap.nShapleyValues): The n-Shapley Values that we want to plot.
        max_degree (int, optional): Plots all effect of order larger than max_degree with a single color. Defaults to None.
        axis (optional): Matplotlib axis on which to plot. Defaults to None.
        feature_names (_type_, optional): Used to label the x-axis. Defaults to None.
        rotation (int, optional): Rotation for x-axis labels. Defaults to 70.
        legend (bool, optional): Plot legend. Defaults to True.
        fig_kwargs (dict, optional): fig_kwargs, handed down to matplotlib figure. Defaults to {"figsize": (6, 6)}.
        barwidth (float, optional): Widht of the bars. Defaults to 0.5.

    Returns:
        Matplotlib axis: The plot axis.
    """
    if max_degree == 1:
        I = I.shapley_values() # TODO this might have to change
    num_features = I.d
    vmax, vmin = 0, 0
    ax = axis
    if axis is None:
        _, ax = plt.subplots(**fig_kwargs)
    if max_degree is None or max_degree >= I.n:
        max_degree = I.n
    ax.axhline(y=0, color="black", linestyle="-")  # line at 0
    for i_feature in range(num_features):
        bmin, bmax = 0, 0
        v = I[(i_feature,)]
        ax.bar(
            x=i_feature,
            height=v,
            width=barwidth,
            bottom=0,
            align="center",
            label=f"Feature {i_feature}",
            color=plot_colors[0],
        )
        bmin = min(bmin, v)
        bmax = max(bmax, v)
        # higher-order effects, up to max_degree
        for n_k in range(2, I.n + 1):
            v_pos = np.sum(
                [
                    I[k] / len(k)
                    for k in I.data.keys()
                    if (len(k) == n_k and i_feature in k and I[k] > 0)
                ]
            )
            v_neg = np.sum(
                [
                    I[k] / len(k)
                    for k in I.data.keys()
                    if (len(k) == n_k and i_feature in k and I[k] < 0)
                ]
            )
            # 'max_degree or higher'
            if n_k == max_degree:
                v_pos = np.sum(
                    [
                        I[k] / len(k)
                        for k in I.data.keys()
                        if (len(k) >= n_k and i_feature in k and I[k] > 0)
                    ]
                )
                v_neg = np.sum(
                    [
                        I[k] / len(k)
                        for k in I.data.keys()
                        if (len(k) >= n_k and i_feature in k and I[k] < 0)
                    ]
                )
            if v_pos > 0:
                ax.bar(
                    x=i_feature,
                    height=v_pos,
                    width=barwidth,
                    bottom=bmax,
                    align="center",
                    color=plot_colors[n_k - 1],
                )
                bmax = bmax + v_pos
            if v_neg < 0:
                ax.bar(
                    x=i_feature,
                    height=v_neg,
                    width=barwidth,
                    bottom=bmin,
                    align="center",
                    color=plot_colors[n_k - 1],
                )
                bmin = bmin + v_neg
            if n_k == max_degree:  # no higher orders
                break
        vmin = min(vmin, bmin)
        vmax = max(vmax, bmax)
    # axes
    if feature_names is None:
        feature_names = [f"Feature {i+1}" for i in range(num_features)]
    ax.set_ylim([1.1 * vmin, 1.1 * vmax])
    ax.set_xticks(np.arange(num_features))
    ax.set_xticklabels(feature_names, rotation=rotation)
    # legend with custom labels
    color_patches = [mpatches.Patch(color=color) for color in plot_colors]
    lables = ["Main"]
    if max_degree > 1:
        lables.append("2nd order")
    if max_degree > 2:
        lables.append(f"3rd order")
        if max_degree > 3:
            for i_degree in range(4, I.n + 1):
                if i_degree == max_degree and max_degree < I.n:
                    lables.append(f"{i_degree}-{I.n}th order")
                    break
                else:
                    if i_degree == I.n:
                        lables.append(f"{i_degree}th order")
                    else:
                        lables.append(f"{i_degree}th")
    ax.legend(
        color_patches,
        lables,
        bbox_to_anchor=(1.02, 1),
        loc="upper left",
        borderaxespad=0,
        handletextpad=0.5,
        handlelength=1,
        handleheight=1,
    )
    ax.get_legend().set_visible(legend)
    ax.set_title(I.index_type)
    if axis is None:
        plt.show()
    return ax



__version__ = "0.2.0"
__author__ = "Sebastian Bordt"

from nshap.InteractionIndex import InteractionIndex
from nshap.InteractionIndex import (
    SHAPLEY_VALUES,
    MOEBIUS_TRANSFORM,
    SHAPLEY_INTERACTION,
    N_SHAPLEY_VALUES,
    SHAPLEY_TAYLOR,
    FAITH_SHAP,
    BANZHAF,
    BANZHAF_INTERACTION,
    FAITH_BANZHAF,
)

from nshap.functions import (
    shapley_interaction_index,
    n_shapley_values,
    moebius_transform,
    shapley_values,
    faith_shap,
    shapley_taylor,
    faith_banzhaf,
    banzhaf_interaction_index,
    bernoulli_numbers,
)
from nshap.plot import plot_interaction_index
from nshap.util import allclose, save, load, powerset

from nshap.vfunc import memoized_vfunc




import xgboost

from test_util import folktables_income

import nshap


def test_n_shapley():
    """Compare different formulas for computing n-shapley values.

    (1) via delta_S
    (2) via the component functions of the shapley gam

    Then compare the shapley values resulting from (1) and (2) to the shapley values computed with the shapley formula.
    """
    X_train, X_test, Y_train, Y_test, _ = folktables_income()
    X_train = X_train[:, 0:5]
    X_test = X_test[:, 0:5]
    gbtree = xgboost.XGBClassifier()
    gbtree.fit(X_train, Y_train)

    vfunc = nshap.vfunc.interventional_shap(gbtree.predict_proba, X_train, target=0)

    n_shapley_values = nshap.n_shapley_values(X_test[0, :], vfunc)
    moebius = nshap.moebius_transform(X_test[0, :], vfunc)
    shapley_values = nshap.shapley_values(X_test[0, :], vfunc)

    assert nshap.allclose(n_shapley_values, moebius)
    for k in range(1, X_train.shape[1]):
        k_shapley_values = nshap.n_shapley_values(X_test[0, :], vfunc, k)
        assert nshap.allclose(n_shapley_values.k_shapley_values(k), k_shapley_values)
        assert nshap.allclose(
            n_shapley_values.k_shapley_values(k), moebius.k_shapley_values(k)
        )
    assert nshap.allclose(n_shapley_values.k_shapley_values(1), shapley_values)
    assert nshap.allclose(moebius.k_shapley_values(1), shapley_values)




import collections

import numpy as np

import nshap

#############################################################################################################################
#                            Unique string identifiers for all supported interaction indices
#############################################################################################################################

SHAPLEY_VALUES = "Shapley Values"
MOEBIUS_TRANSFORM = "Moebius Transform"
SHAPLEY_INTERACTION = "Shapley Interaction"
N_SHAPLEY_VALUES = "n-Shapley Values"
SHAPLEY_TAYLOR = "Shapley Taylor"
FAITH_SHAP = "Faith-Shap"
BANZHAF = "Banzhaf"
BANZHAF_INTERACTION = "Banzhaf Interaction"
FAITH_BANZHAF = "Faith-Banzhaf"

ALL_INDICES = [
    SHAPLEY_VALUES,
    MOEBIUS_TRANSFORM,
    SHAPLEY_INTERACTION,
    N_SHAPLEY_VALUES,
    SHAPLEY_TAYLOR,
    FAITH_SHAP,
    BANZHAF,
    BANZHAF_INTERACTION,
    FAITH_BANZHAF
]

#############################################################################################################################
#                                  A single class for all interaction indices
#############################################################################################################################


class InteractionIndex(collections.UserDict):
    """Class for different interaction indices (n-Shapley Values, Shapley Taylor, Faith-Shap).
    Interaction indices are a Python dict with added functionality.
    """

    def __init__(self, index_type: str, values, n=None, d=None):
        """Initialize an interaction index from a dict of values.

        Args:
            type (str): The type of the interaction index.
            values (_type_): The underlying dict of values.
            n (_type_, optional): _description_. Defaults to None.
            d (_type_, optional): _description_. Defaults to None.
        """
        super().__init__(values)
        assert index_type in ALL_INDICES, f"{index_type} is not a supported interaction index"
        self.index_type = index_type
        self.n = n
        if (
            n is None
        ):  # if n or d are not given as aruments, infer them from the values dict.
            self.n = max([len(x) for x in values.keys()])
        self.d = d
        if d is None:
            self.d = len([x[0] for x in values.keys() if len(x) == 1])

    def get_index_type(self):
        """Return the type of the interaction index (for example, "Shapley Taylor")

        Returns:
            str: See function description.
        """
        return self.index_type

    def get_input_dimension(self):
        """Return the input dimension of the function for which we computed the interaction index  ('d').

        Returns:
            integer: See function description.
        """
        return self.d

    def get_order(self):
        """Return the order of the interaction index ('n').

        Returns:
            integer: See function description.
        """
        return self.n

    def sum(self):
        """Sum all the terms that are invovled in the interaction index.

        For many interaction indices, the result is equal to the value of the function that we attempt to explain, minus the value of the empty coalition.

        Returns:
            Float: The sum.
        """
        return np.sum([x for x in self.data.values()])

    def copy(self):
        """Return a copy of the current object.

        Returns:
            nshap.InteractionIndex: The copy.
        """
        return InteractionIndex(self.index_type, self.data.copy())

    def save(self, fname):
        """Save the interaction index to a JSON file.

        Args:
            fname (str): Filename.
        """
        nshap.save(self, fname)

    def plot(self, *args, **kwargs):
        """Generate a plots of the n-Shapley Values.

        This function simply calls nshap.plots.plot_n_shapley.

        Returns:
            The axis of the matplotlib plot.
        """
        return nshap.plot_interaction_index(self, *args, **kwargs)

    def shapley_values(self):
        """Compute the original Shapley Values.

        Returns:
            numpy.ndarray: Shaley Values. If you prefer an object of type nShapleyValues, call k_shapley_values(self, 1).
        """
        assert (
            self.index_type == N_SHAPLEY_VALUES or self.index_type == MOEBIUS_TRANSFORM
        ), f"shapley_values only supports {N_SHAPLEY_VALUES} and {MOEBIUS_TRANSFORM}"
        shapley_values = self.k_shapley_values(1)
        shapley_values = np.array(list(shapley_values.values())).reshape((1, -1))
        return shapley_values

    def k_shapley_values(self, k):
        """Compute k-Shapley Values of lower order. Requires k <= n.

        Args:
            k (int): The desired order.

        Returns:
            nShapleyValues: k-Shapley Values.
        """
        assert (
            self.index_type == N_SHAPLEY_VALUES or self.index_type == MOEBIUS_TRANSFORM
        ), f"k_shapley_values only supports {N_SHAPLEY_VALUES} and {MOEBIUS_TRANSFORM}"
        assert k <= self.n, "k_shapley_values requires k<n"
        result = self.copy()
        result.index_type = N_SHAPLEY_VALUES
        for _ in range(k, self.n):
            result = result._n_minus_one_values()
        return result

    def _n_minus_one_values(self):
        """Compute the k-Shapley Values for k=n-1.

        Returns:
            nShapleyValues: (n-1)-Shapley Values.
        """
        assert (
            self.index_type == N_SHAPLEY_VALUES or self.index_type == MOEBIUS_TRANSFORM
        ), f"_n_minus_one_values only supports {N_SHAPLEY_VALUES} and {MOEBIUS_TRANSFORM}"
        result = {}
        # consider all subsets S with 1<=|S|<=n-1
        for S in nshap.powerset(range(self.d)):
            if (len(S) == 0) or (len(S) > self.n - 1):
                continue
            # we have the n-normalized effect
            S_effect = self.data.get(S, 0)
            # go over all subsets T of length n that contain S
            for T in nshap.powerset(range(self.d)):
                if (len(T) != self.n) or (not set(S).issubset(T)):
                    continue
                # add the effect of T to S
                T_effect = self.data.get(
                    T, 0
                )  # default to zero in case the dict is sparse
                # normalization
                S_effect = (
                    S_effect - (nshap.bernoulli_numbers[len(T) - len(S)]) * T_effect
                )
            # now we have the normalized effect
            result[S] = S_effect
        return InteractionIndex(N_SHAPLEY_VALUES, result)



import numpy as np
import math

import nshap


#############################################################################################################################
#                                                   Bernoulli numbers
#############################################################################################################################

bernoulli_numbers = np.array(
    [
        1,
        -1 / 2,
        1 / 6,
        0,
        -1 / 30,
        0,
        1 / 42,
        0,
        -1 / 30,
        0,
        5 / 66,
        0,
        -691 / 2730,
        0,
        7 / 6,
        0,
        -3617 / 510,
        0,
        43867 / 798,
        0,
    ]
)


#############################################################################################################################
#                                   Input Validation (Used by all that follows)
#############################################################################################################################


def validate_inputs(x, v_func):
    if x.ndim == 1:
        x = x.reshape((1, -1))
    assert (
        x.shape[0] == 1
    ), "The nshap package only accepts single data points as input."
    dim = x.shape[1]
    if not isinstance(v_func, nshap.memoized_vfunc):  # meomization
        v_func = nshap.memoized_vfunc(v_func)
    # for d>20, we would have to consider the numerics of the problem more carefully
    assert dim <= 20, "The nshap package only supports d<=20."
    return x, v_func, dim


#############################################################################################################################
#                                               The Moebius Transform
#############################################################################################################################


def moebius_transform(x, v_func):
    """Compute the Moebius Transform of of the value function v_func at the data point x.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    # go over all subsets S of N with 1<=|S|<=d
    result = {}
    for S in nshap.powerset(set(range(dim))):
        if len(S) == 0:
            continue
        summands = []
        # go over all subsets T of S
        for T in nshap.powerset(S):
            summands.append(v_func(x, list(T)) * (-1) ** (len(S) - len(T)))
        result[S] = np.sum(summands)
    # return result
    return nshap.InteractionIndex(nshap.MOEBIUS_TRANSFORM, result)


#############################################################################################################################
#                                                   Shapley Values
#############################################################################################################################


def shapley_values(x, v_func):
    """Compute the original Shapley Values, according to the Shapley Formula.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    # go over all features
    result = {}
    for i_feature in range(dim):
        phi = 0
        S = set(range(dim))
        S.remove(i_feature)
        for subset in nshap.powerset(S):
            v = v_func(x, list(subset))
            subset_i = list(subset)
            subset_i.append(i_feature)
            subset_i.sort()
            v_i = v_func(x, subset_i)
            phi = phi + np.math.factorial(len(subset)) * np.math.factorial(
                dim - len(subset) - 1
            ) / np.math.factorial(dim) * (v_i - v)
        result[(i_feature,)] = phi
    # return result
    return nshap.InteractionIndex(nshap.SHAPLEY_VALUES, result)


#############################################################################################################################
#                                           Shapley Interaction Index
#############################################################################################################################


def shapley_interaction_index(x, v_func, n=-1):
    """Compute the Shapley Interaction Index (https://link.springer.com/article/10.1007/s001820050125) at the data point x, and all S such that |S|<=n, given a coalition value function.

    Args:
        x (numpy.ndarray): A data point
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int): Order up to which the Shapley Interaction Index should be computed.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # go over all subsets S of N with |S|<=n
    result = {}
    for S in nshap.powerset(set(range(dim))):
        if len(S) > n:
            continue
        # go over all subsets T of N\S
        phi = 0
        N_minus_S = set(range(dim)) - set(S)
        for T in nshap.powerset(N_minus_S):
            # go over all subsets L of S
            delta = 0
            for L in nshap.powerset(S):
                coalition = list(L)
                coalition.extend(list(T))
                coalition.sort()
                delta = delta + np.power(-1, len(S) - len(L)) * v_func(x, coalition)
            phi = phi + delta * np.math.factorial(len(T)) * np.math.factorial(
                dim - len(T) - len(S)
            ) / np.math.factorial(dim - len(S) + 1)
        result[S] = phi
    # return result
    return nshap.InteractionIndex(nshap.SHAPLEY_INTERACTION, result)


#############################################################################################################################
#                                                 n-Shapley Values
#############################################################################################################################


def n_shapley_values(x, v_func, n=-1):
    """This function provides an exact computation of n-Shapley Values (https://arxiv.org/abs/2209.04012) via their definition.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int, optional): Order of n-Shapley Values or -1 for n=d. Defaults to -1.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # first compute the shapley interaction index
    shapley_int_idx = shapley_interaction_index(x, v_func, n)
    # a list of length num_datapoints
    result = {}
    # consider all subsets S with 1<=|S|<=n
    for S in nshap.powerset(range(dim)):
        if (len(S) == 0) or (len(S) > n):
            continue
        # obtain the unnormalized effect (that is, delta_S(x))
        S_effect = shapley_int_idx[S]
        # go over all subsets T of length k+1, ..., n that contain S
        for T in nshap.powerset(range(dim)):
            if (len(T) <= len(S)) or (len(T) > n) or (not set(S).issubset(T)):
                continue
            # get the effect of T, and substract it from the effect of S
            T_effect = shapley_int_idx[T]
            # normalization with bernoulli_numbers
            S_effect = S_effect + (bernoulli_numbers[len(T) - len(S)]) * T_effect
        # now we have the normalized effect
        result[S] = S_effect
    # return result
    return nshap.InteractionIndex(nshap.N_SHAPLEY_VALUES, result)


#############################################################################################################################
#                                           Shapley Taylor Interaction Index
#############################################################################################################################


def shapley_taylor(x, v_func, n=-1):
    """ Compute the Shapley Taylor Interaction Index (https://arxiv.org/abs/1902.05622) of the value function v_func at the data point x.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int, optional): Order of the Shapley Taylor Interaction Index or -1 for n=d. Defaults to -1.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # we first compute the moebius transform
    moebius = moebius_transform(x, v_func)
    # then compute the Shapley Taylor Interaction Index, for all datapoints
    result = {}
    # consider all subsets S with 1<=|S|<=n
    for S in nshap.powerset(range(dim)):
        if (len(S) == 0) or (len(S) > n):
            continue
        result[S] = moebius[S]
        # for |S|=n, average the higher-order effects
        if len(S) == n:
            # go over all subsets of [d] that contain S
            for T in nshap.powerset(range(dim)):
                if (len(T) <= len(S)) or (not set(S).issubset(T)):
                    continue
                result[S] += moebius[T] / math.comb(len(T), len(S))
    # return result
    return nshap.InteractionIndex(nshap.SHAPLEY_TAYLOR, result)


#############################################################################################################################
#                                             Faith-Shap Interaction Index
#############################################################################################################################


def faith_shap(x, v_func, n=-1):
    """ Compute the Faith-Shap Interaction Index (https://arxiv.org/abs/2203.00870) of the value function v_func at the data point x.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int, optional): Order of the Interaction Index or -1 for n=d. Defaults to -1.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # we first compute the moebius transform
    moebius = moebius_transform(x, v_func)
    # then compute the Faith-Shap Interaction Index
    result = {}
    # consider all subsets S with 1<=|S|<=n
    for S in nshap.powerset(range(dim)):
        if (len(S) == 0) or (len(S) > n):
            continue
        result[S] = moebius[S]
        # go over all subsets of [d] that contain S
        for T in nshap.powerset(range(dim)):
            if (len(T) <= n) or (not set(S).issubset(T)):
                continue
            # compare Theorem 19 in the Faith-Shap paper. In our notation, l=n.
            result[S] += (
                (-1) ** (n - len(S))
                * len(S)
                / (n + len(S))
                * math.comb(n, len(S))
                * math.comb(len(T) - 1, n)
                / math.comb(len(T) + n - 1, n + len(S))
                * moebius[T]
            )
    # return result
    return nshap.InteractionIndex(nshap.FAITH_SHAP, result)


#############################################################################################################################
#                                                 Bhanzaf Values
#############################################################################################################################


#############################################################################################################################
#                                           Bhanzaf Interaction Index
#############################################################################################################################


def banzhaf_interaction_index(x, v_func, n=-1):
    """Compute the Banzhaf Interaction Index (https://link.springer.com/article/10.1007/s001820050125) at the data point x, and all S such that |S|<=n, given a coalition value function.

    Args:
        x (numpy.ndarray): A data point
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int): Order up to which the Shapley Interaction Index should be computed.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # go over all subsets S of N with |S|<=n
    result = {}
    for S in nshap.powerset(set(range(dim))):
        if len(S) > n:
            continue
        # go over all subsets T of N\S
        phi = 0
        N_minus_S = set(range(dim)) - set(S)
        for T in nshap.powerset(N_minus_S):
            # go over all subsets L of S
            delta = 0
            for L in nshap.powerset(S):
                coalition = list(L)
                coalition.extend(list(T))
                coalition.sort()
                delta = delta + np.power(-1, len(S) - len(L)) * v_func(x, coalition)
            phi = phi + delta * (1 / np.power(2, dim - len(S)))
        result[S] = phi
    # return result
    return nshap.InteractionIndex(nshap.BANZHAF_INTERACTION, result)


#############################################################################################################################
#                                         Faith-Bhanzaf Interaction Index
#############################################################################################################################


def faith_banzhaf(x, v_func, n=-1):
    """ Compute the Faith-Banzhaf Interaction Index (https://arxiv.org/abs/2203.00870) of the value function v_func at the data point x.

    Args:
        x (numpy.ndarray): A data point.
        v_func (function): The value function. It takes two arguments: The datapoint x and a list with the indices of the coalition.
        n (int, optional): Order of the Interaction Index or -1 for n=d. Defaults to -1.

    Returns:
        nshap.InteractionIndex: The interaction index.
    """
    # validate input parameters
    x, v_func, dim = validate_inputs(x, v_func)
    if n == -1:
        n = dim
    # we first compute the moebius transform
    moebius = moebius_transform(x, v_func)
    # then compute the Faith-Banzhaf Interaction Index
    result = {}
    # consider all subsets S with 1<=|S|<=n
    for S in nshap.powerset(range(dim)):
        if (len(S) == 0) or (len(S) > n):
            continue
        result[S] = moebius[S]
        # go over all subsets of [d] that contain S
        for T in nshap.powerset(range(dim)):
            if (len(T) <= n) or (not set(S).issubset(T)):
                continue
            # compare Theorem 17 in the Faith-Shap paper. In our notation, l=n.
            result[S] += (
                (-1) ** (n - len(S))
                * np.power(0.5, len(T) - len(S))
                * math.comb(len(T) - len(S) - 1, n - len(S))
                * moebius[T]
            )
    # return result
    return nshap.InteractionIndex(nshap.FAITH_BANZHAF, result)



import xgboost

from folktables import ACSDataSource, ACSIncome

from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

import os

data_root_dir = "../data/"

paths = [data_root_dir]
for p in paths:
    if not os.path.exists(p):
        os.mkdir(p)


def folktables_income():
    # (down-)load the dataset
    data_source = ACSDataSource(
        survey_year="2016", horizon="1-Year", survey="person", root_dir=data_root_dir
    )
    data = data_source.get_data(states=["CA"], download=True)
    X, Y, _ = ACSIncome.df_to_numpy(data)
    feature_names = ACSIncome.features

    # feature names
    feature_names = ACSIncome.features

    # zero mean and unit variance for all features
    X = StandardScaler().fit_transform(X)

    # train-test split
    X_train, X_test, Y_train, Y_test = train_test_split(
        X, Y, train_size=0.8, random_state=0
    )

    return X_train, X_test, Y_train, Y_test, feature_names



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nshap",
    version="0.2.0",
    author="Sebastian Bordt",
    author_email="sbordt@posteo.de",
    description="Python package to compute n-Shapley Values.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tml-tuebingen/nshap",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=["nshap"],
    python_requires=">=3.6",
    install_requires=["numpy", "matplotlib", "seaborn",],
)



import numpy as np

import functools

import numbers

#############################################################################################################################
#                                                Interventional SHAP
#############################################################################################################################


def interventional_shap(
    f, X, target=None, num_samples=1000, random_state=None, meomized=True
):
    """Approximate the value function of interventional SHAP.

    In order to compute n-Shapley Values, a data set is sampled once and then fixed for all evaluations of the value function.

    Args:
        f (function): The function to be explained. Will be called as f(x) where x has shape (1,d).
        X (_type_): Sample from the data distribution.
        target (int, optional): Target class. Required if the output of f(x) is multi-dimensional. Defaults to None.
        num_samples (int, optional): The number of samples  that should be drawn from X in order estimate the value function. Defaults to 1000.
        random_state (_type_, optional): Random state that is passed to np.random.default_rng. Used for reproducibility. Defaults to None.
        meomized (bool, optional): Whether the returned value function should be meomized. Defaults to True.

    Returns:
        function: The vaue function.
    """
    # sample background data set
    rng = np.random.default_rng(random_state)
    if num_samples < X.shape[0]:
        indices = rng.integers(low=0, high=X.shape[0], size=num_samples)
        X = X[indices, :]
    # the function
    def fn(x, subset):
        subset = list(subset)
        x = x.flatten()
        values = []
        for idx in range(X.shape[0]):
            x_sample = X[idx, :].copy()  # copy here is important, otherwise we modify X
            x_sample[subset] = x[subset]
            v = f(x_sample.reshape((1, -1)))
            # handle different return types
            if isinstance(v, numbers.Number):
                values.append(v)
            elif v.ndim == 1:
                if target is None:
                    values.append(v[0])
                else:
                    values.append(v[target])
            else:
                assert (
                    target is not None
                ), "f returns multi-dimensional array, but target is not specified"
                values.append(v[0, target])
        return np.mean(values)

    if meomized:  # meomization
        fn = memoized_vfunc(fn)
    return fn


#############################################################################################################################
#                                          Meomization for value functions
#############################################################################################################################


class memoized_vfunc(object):
    """Decorator to meomize a vfunc. Handles hashability of the numpy.ndarray and the list.

    The hash depends only on the values of x that are in the subset of coordinates.

    This function is able to handle parameters x of shape (d,) and (1,d).

    Meomization helps to avoid duplicate evaluations of the value function, even if we perform
    computations that involve the same term multiple times.

    https://wiki.python.org/moin/PythonDecoratorLibrary
    """

    def __init__(self, func):
        """

        Args:
            func (function): The value function to be meomized.
        """
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        x = args[0]
        if x.ndim == 2:
            assert (
                x.shape[0] == 1
            ), "Parameter x of value function has to be of shape (d,) or (1,d)."
            x = x[0]
        subset = tuple(args[1])
        x = tuple((x[i] for i in subset))
        hashable_args = (x, subset)
        if hashable_args in self.cache:  # in cache
            return self.cache[hashable_args]
        else:
            value = self.func(*args)
            self.cache[hashable_args] = value
            return value

    def __repr__(self):
        """Return the function's docstring."""
        return f"Memoized value function: {self.func.__doc__}"

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)
